from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional

from ..writer import LogWriter
from .base import BaseContainer
from .types import Entity


@dataclass
class RetrievalConfig():
    id: str
    name: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


class Retrieval(BaseContainer):
    def __init__(self, config: RetrievalConfig, writer: LogWriter):
        super().__init__(Entity.RETRIEVAL, config.__dict__, writer)
        self._commit("create")

    def input(self, query: str):
        self._commit("update", {"input": query})
        self.end()

    @staticmethod
    def input_(writer: LogWriter, id: str, query: str):
        BaseContainer._commit_(writer, Entity.RETRIEVAL,
                               id, "update", {"input": query})

    def output(self, docs: Any):
        self._commit("output", {"docs": docs, "endTimestamp": datetime.now()})
        self.end()

    @staticmethod
    def output_(writer: LogWriter, id: str, docs: Any):
        BaseContainer._commit_(writer, Entity.RETRIEVAL, id, "output", {
                               "docs": docs})
        BaseContainer._end_(writer, Entity.RETRIEVAL, id, {
                            "endTimestamp": datetime.now()})

    @staticmethod
    def end_(writer: LogWriter, id: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = {}
        BaseContainer._end_(writer, Entity.RETRIEVAL, id, {
            "endTimestamp": datetime.now(),
            **data,
        })

    @staticmethod
    def add_tag_(writer: LogWriter, id: str, key: str, value: str):
        BaseContainer._add_tag_(writer, Entity.RETRIEVAL, id, key, value)



import json
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Dict, List, Optional

from ..writer import LogWriter
from .base import BaseContainer
from .types import Entity


@dataclass
class GenerationConfig:
    id: str
    provider: str
    model: str
    messages: List[Any] = field(default_factory=list)
    model_parameters: Dict[str, Any] = field(default_factory=dict)
    span_id: Optional[str] = None
    name: Optional[str] = None
    maxim_prompt_id: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


@dataclass
class Choice:
    index: int
    text: str
    logprobs: Optional[object] = None
    finish_reason: Optional[str] = None


@dataclass
class Usage:
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


@dataclass
class TextCompletion:
    id: str
    object: str
    created: int
    model: str
    choices: List[Choice]
    usage: Usage

# Validators


def validate_type(value, expected_type, field_name):
    if not isinstance(value, expected_type):
        raise ValueError(
            f"{field_name} must be of type {expected_type.__name__}")


def validate_choice(choice_data):
    validate_type(choice_data.get('index'), int, 'index')
    validate_type(choice_data.get('text'), str, 'text')
    validate_type(choice_data.get('finish_reason'), str, 'finish_reason')
    return Choice(**choice_data)


def validate_usage(usage_data):
    validate_type(usage_data.get('prompt_tokens'), int, 'prompt_tokens')
    validate_type(usage_data.get('completion_tokens'),
                  int, 'completion_tokens')
    validate_type(usage_data.get('total_tokens'), int, 'total_tokens')
    return Usage(**usage_data)


def validate_text_completion(data: dict) -> TextCompletion:
    validate_type(data.get('id'), str, 'id')
    validate_type(data.get('object'), str, 'object')
    validate_type(data.get('created'), int, 'created')
    validate_type(data.get('model'), str, 'model')

    choices_data = data.get('choices')
    validate_type(choices_data, list, 'choices')
    if not choices_data:
        raise ValueError("choices must not be empty")
    choices = [validate_choice(choice) for choice in choices_data]

    usage = validate_usage(data.get('usage', {}))

    return TextCompletion(
        id=data['id'],
        object=data['object'],
        created=data['created'],
        model=data['model'],
        choices=choices,
        usage=usage
    )


# Main class
class Generation(BaseContainer):
    def __init__(self, config: GenerationConfig, writer: LogWriter):
        super().__init__(Entity.GENERATION, config.__dict__, writer)
        self.model = config.model
        self.maxim_prompt_id = config.maxim_prompt_id
        self.messages = []
        self.provider = config.provider
        self.messages.extend(config.messages)
        self.model_parameters = config.model_parameters

    @staticmethod
    def set_model_(writer: LogWriter, id: str, model: str):
        BaseContainer._commit_(writer, Entity.GENERATION,
                               id, "update", {"model": model})

    def set_model(self, model: str):
        self.model = model
        self._commit("update", {"model": model})

    @staticmethod
    def add_message_(writer: LogWriter, id: str, message: Any):
        BaseContainer._commit_(writer, Entity.GENERATION, id, "update", {
            "messages": [message]})

    def add_message(self, message: Any):
        self.messages.append(message)
        self._commit("update", {"messages": [message]})

    @staticmethod
    def set_model_parameters_(writer: LogWriter, id: str, model_parameters: Dict[str, Any]):
        BaseContainer._commit_(writer, Entity.GENERATION, id, "update", {
            "model_parameters": model_parameters})

    def set_model_parameters(self, model_parameters: Dict[str, Any]):
        self.model_parameters = model_parameters
        self._commit("update", {"model_parameters": model_parameters})

    @staticmethod
    def result_(writer: LogWriter, id: str, result: Any):
        try:
            validate_text_completion(result)
            BaseContainer._commit_(writer,
                                   Entity.GENERATION, id, "result", {"result": result})
            BaseContainer._end_(writer, Entity.GENERATION, id, {
                "endTimestamp": datetime.now(),
            })
        except ValueError as e:
            raise ValueError(f"Invalid result. We expect OpenAI response format: {e}")

    @staticmethod
    def end_(writer: LogWriter, id: str, data: Optional[Dict[str, Any]] = None):
        if data is None:
            data = {}
        BaseContainer._end_(writer, Entity.GENERATION, id, {
            "endTimestamp": datetime.now(),
            **data,
        })

    @staticmethod
    def add_tag_(writer: LogWriter, id: str, key: str, value: str):
        BaseContainer._add_tag_(writer, Entity.GENERATION, id, key, value)

    def result(self, result: Any):
        try:
            validate_text_completion(result)
            self._commit("result", {"result": result})
            self.end()
        except ValueError as e:
            raise ValueError(f"Invalid result. We expect OpenAI response format: {e}")
        
    def error(self, error:str):
        self._commit("error", {"error": error})
        self.end()
    
    @staticmethod
    def error_(writer: LogWriter, id: str, error: str):
        BaseContainer._commit_(writer, Entity.GENERATION, id, "error", {"error": error})
        BaseContainer._end_(writer, Entity.GENERATION, id, {
            "endTimestamp": datetime.now(),
        })

    def data(self) -> Dict[str, Any]:
        base_data = super().data()
        return {
            **base_data,
            "model": self.model,
            "provider": self.provider,
            "maximPromptId": self.maxim_prompt_id,
            "messages": self.messages,
            "modelParameters": self.model_parameters,
        }



import json
import logging
import threading
import time
from dataclasses import asdict, dataclass
from typing import Dict, List, Optional, Union

from .apis.maxim import MaximAPI
from .cache.cache import MaximCache
from .cache.inMemory import MaximInMemoryCache
from .filterObjects import (IncomingQuery, QueryObject, findAllMatches,
                            findBestMatch, parseIncomingQuery)
from .logger.logger import Logger, LoggerConfig
from .models.dataset import DatasetEntry
from .models.folder import Folder, FolderEncoder
from .models.prompt import (Prompt, PromptData, PromptVersion, RuleGroupType,
                            RuleType, VersionAndRulesWithPromptIdEncoder,
                            VersionsAndRules, VersionSpecificDeploymentConfig)
from .models.queryBuilder import QueryRule

logging.basicConfig(level=logging.ERROR)


@dataclass
class Config():
    apiKey: str
    baseUrl: Optional[str] = None
    cache: Optional[MaximCache] = None
    debug: Optional[bool] = False


EntityType = {
    "PROMPT": "PROMPT",
    "FOLDER": "FOLDER"
}


class Interval:
    def __init__(self, interval, action):
        self.interval = interval
        self.action = action
        self.stopEvent = threading.Event()
        thread = threading.Thread(target=self.__setInterval)
        thread.start()

    def __setInterval(self):
        nextTime = time.time()+self.interval
        while not self.stopEvent.wait(nextTime-time.time()):
            nextTime += self.interval
            self.action()

    def cancel(self):
        self.stopEvent.set()


class Maxim:
    def __init__(self, config: Config):
        self.baseUrl = config.baseUrl or "https://app.getmaxim.ai"
        self.apiKey = config.apiKey
        self.__isDebug = config.debug or False
        self.__cache = config.cache or MaximInMemoryCache()
        self.__sync = self.__syncEntities()
        self.__loggers: Dict[str, Logger] = {}
        self.__syncInterval = Interval(60, self.__syncEntities)

    def __syncEntities(self):
        self.__syncPrompts()
        self.__syncFolders()

    def __syncPrompts(self):
        prompts = MaximAPI.getPrompts(self.baseUrl, self.apiKey)
        for prompt in prompts:
            try:
                self.__cache.set(self.__getCacheKey(
                    EntityType['PROMPT'], prompt.promptId), json.dumps(prompt, cls=VersionAndRulesWithPromptIdEncoder))
            except Exception as err:
                raise Exception(err)

    def __syncFolders(self):
        folders = MaximAPI.getFolders(self.baseUrl, self.apiKey)
        for folder in folders:
            try:
                self.__cache.set(self.__getCacheKey(
                    EntityType['FOLDER'], folder.id), json.dumps(folder, cls=FolderEncoder))
            except Exception as err:
                raise Exception(err)

    def __getCacheKey(self, entity: str, id: str) -> str:
        if entity == 'PROMPT':
            return f"prompt:{id}"
        else:
            return f"folder:{id}"

    def __getPromptFromCache(self, key: str) -> Optional[VersionsAndRules]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        if 'promptId' in json_data:
            json_data.pop('promptId')
        return VersionsAndRules.from_dict(json_data)

    def __getAllPromptsFromCache(self) -> Optional[List[VersionsAndRules]]:
        keys = self.__cache.getAllKeys()
        if not keys:
            return None
        data = [self.__cache.get(key)
                for key in keys if key.startswith("prompt:")]
        promptList = []
        for d in data:
            if d is not None:
                json_data = json.loads(d)
                if 'promptId' in json_data:
                    json_data.pop('promptId')
                promptList.append(VersionsAndRules.from_dict(json_data))
        return promptList

    def __getFolderFromCache(self, key: str) -> Optional[Folder]:
        data = self.__cache.get(key)
        if not data:
            return None
        json_data = json.loads(data)
        return Folder(**json_data)

    def __getAllFoldersFromCache(self) -> Optional[List[Folder]]:
        keys = self.__cache.getAllKeys()
        if not keys:
            return None
        data = [self.__cache.get(key)
                for key in keys if key.startswith("folder:")]
        return [Folder(**json.loads(d)) for d in data if d is not None]

    def __formatPrompt(self, prompt: PromptVersion) -> Prompt:
        json_data = {
            'promptId': prompt.promptId,
            'version': prompt.version,
            'versionId': prompt.id,
            'messages': prompt.config.messages if prompt.config else None,
            'modelParameters': prompt.config.modelParameters if prompt.config else None
        }
        return Prompt(**json_data)

    def __getPromptVersionForRule(self, promptVersionAndRules: VersionsAndRules, rule: Optional[QueryRule] = None) -> Optional[Prompt]:
        if rule:
            incomingQuery = IncomingQuery(
                query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch)
            objects = []
            for versionId, versionRules in promptVersionAndRules.rules.items():
                for versionRule in versionRules:
                    if not versionRule.rules.query:
                        continue
                    if rule.scopes:
                        for key in rule.scopes.keys():
                            if key != "folder":
                                raise ValueError("Invalid scope added")
                    version = next(
                        (v for v in promptVersionAndRules.versions if v.id == versionId), None)
                    if not version:
                        continue
                    query = versionRule.rules.query
                    if version.config and version.config.tags:
                        parsedIncomingQuery = parseIncomingQuery(
                            incomingQuery.query)
                        tags = version.config.tags
                        for key, value in tags.items():
                            if value == None:
                                continue
                            if parsedIncomingQuery is None:
                                continue
                            incomingQueryFields = [
                                q.field for q in parsedIncomingQuery]
                            if key in incomingQueryFields:
                                query.rules.append(
                                    RuleType(field=key, operator='=', value=value))
                    objects.append(QueryObject(query=query, id=versionId))

            deployedVersionObject = findBestMatch(objects, incomingQuery)
            if deployedVersionObject:
                deployedVersion = next(
                    (v for v in promptVersionAndRules.versions if v.id == deployedVersionObject.id), None)
                if deployedVersion:
                    return self.__formatPrompt(deployedVersion)

        else:
            if promptVersionAndRules.rules:
                for versionId, versionRules in promptVersionAndRules.rules.items():
                    logging.debug(f"type: {type(versionRules[0].rules)}")
                    doesQueryExist = any(
                        ruleElm.rules.query != None for ruleElm in versionRules)
                    if doesQueryExist:
                        deployedVersion = next(
                            (v for v in promptVersionAndRules.versions if v.id == versionId), None)
                        if deployedVersion:
                            return self.__formatPrompt(deployedVersion)
            else:
                return self.__formatPrompt(promptVersionAndRules.versions[0])

        if promptVersionAndRules.fallbackVersion:
            print(
                f"************* FALLBACK TRIGGERED : VERSION : {promptVersionAndRules.fallbackVersion} *************")
            return self.__formatPrompt(promptVersionAndRules.fallbackVersion)
        return None

    def __getFoldersForRule(self, folders: List[Folder], rule: QueryRule) -> List[Folder]:
        incomingQuery = IncomingQuery(
            query=rule.query, operator=rule.operator, exactMatch=rule.exactMatch)
        objects = []
        for folder in folders:
            query = RuleGroupType(rules=[], combinator='AND')
            if not folder.tags:
                continue
            parsedIncomingQuery = parseIncomingQuery(incomingQuery.query)
            tags = folder.tags
            for key, value in tags.items():
                if key in [q.field for q in parsedIncomingQuery]:
                    # if not isinstance(value,None):
                    query.rules.append(
                        RuleType(field=key, operator='=', value=value))
            if not query.rules:
                continue
            objects.append(QueryObject(query=query, id=folder.id))

        folderObjects = findAllMatches(objects, incomingQuery)
        ids = [fo.id for fo in folderObjects]
        return [f for f in folders if f.id in ids]

    def getPrompt(self, id: str, rule: Optional[QueryRule] = None) -> Prompt:
        key = self.__getCacheKey('PROMPT', id)
        versionAndRules = self.__getPromptFromCache(key)
        if versionAndRules is None:
            versionAndRules = MaximAPI.getPrompt(self.baseUrl, self.apiKey, id)
            if len(versionAndRules.versions) == 0:
                raise Exception(f'No active deployments found for Prompt {id}')
            self.__cache.set(id, json.dumps(versionAndRules))
        if not versionAndRules:
            raise Exception(f'No active deployments found for Prompt {id}')
        prompt = self.__getPromptVersionForRule(versionAndRules, rule)
        if not prompt:
            raise Exception(f'No active deployments found for Prompt {id}')
        return prompt

    def getPrompts(self, rule: QueryRule) -> List[Prompt]:
        versionAndRules = self.__getAllPromptsFromCache()
        prompts = []
        if versionAndRules is None or len(versionAndRules) == 0:
            self.__syncEntities()
            versionAndRules = self.__getAllPromptsFromCache()
        if not versionAndRules:
            raise Exception('No active deployments found for any prompt')
        for v in versionAndRules:
            if rule.scopes:
                if 'folder' not in rule.scopes.keys():
                    raise ValueError("Invalid scope added")
                else:
                    if rule.scopes['folder'] != v.folderId:
                        continue
            prompt = self.__getPromptVersionForRule(v, rule)
            if prompt != None:
                prompts.append(prompt)
        if len(prompts) == 0:
            raise Exception('No active deployments found for any prompt')
        return prompts

    def getFolderById(self, id: str) -> Folder:
        key = self.__getCacheKey('FOLDER', id)
        folder = self.__getFolderFromCache(key)
        if folder is None:
            folder = MaximAPI.getFolder(self.baseUrl, self.apiKey, id)
            if not folder:
                raise Exception(f'No folder found with id {id}')
            self.__cache.set(key, json.dumps(folder))
        return folder

    def getFolders(self, rule: QueryRule) -> List[Folder]:
        folders = self.__getAllFoldersFromCache()
        if folders is None or len(folders) == 0:
            self.__syncEntities()
            folders = self.__getAllFoldersFromCache()
        if not folders:
            raise Exception('No folders found')
        return self.__getFoldersForRule(folders, rule)

    def addDatasetEntries(self, datasetId: str, entries):
        entriesList: List[DatasetEntry] = []
        for entry in entries:
            try:
                entriesList.append(DatasetEntry(**entry))
            except Exception as e:
                raise ValueError(
                    f'couldn\'t cast {entry} to DatasetEntry object. Please adhere to the format of DataseteEntry.')
        MaximAPI.addDatasetEntries(
            self.baseUrl, self.apiKey, datasetId, entriesList)

    def logger(self, config: LoggerConfig):
        # Checking if this log repository exist on server
        exists = MaximAPI.doesLogRepositoryExist(
            self.baseUrl, self.apiKey, config.id)
        if not exists:
            if config.id:
                raise Exception("Log repository not found")
        if config.id in self.__loggers:
            return self.__loggers[config.id]
        logger = Logger(config=config, api_key=self.apiKey,
                        base_url=self.baseUrl, is_debug=self.__isDebug)
        self.__loggers[config.id] = logger
        return logger

    def cleanup(self):
        self.__syncInterval.cancel()
        # Cleaning up all loggers
        for logger in self.__loggers.values():
            logger.cleanup()



from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Dict, Optional

from ..writer import LogWriter
from .base import EventEmittingBaseContainer
from .feedback import Feedback
from .generation import Generation, GenerationConfig
from .retrieval import Retrieval, RetrievalConfig
from .types import Entity

if TYPE_CHECKING:
    from .span import Span, SpanConfig  # Type checking only


@dataclass
class TraceConfig:
    id: str
    name: Optional[str] = None
    session_id: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


class Trace(EventEmittingBaseContainer):
    def __init__(self, config: TraceConfig, writer: LogWriter):
        super().__init__(Entity.TRACE, config.__dict__, writer)
        self._commit("create", {
            **self.data(),
            "sessionId": config.session_id,
        })

    def generation(self, config: GenerationConfig) -> Generation:
        generation = Generation(config, self.writer)
        self._commit("add-generation", {
            **generation.data(),
            "id": generation.id,
        })
        return generation

    @staticmethod
    def generation_(writer: LogWriter, trace_id: str, config: GenerationConfig) -> Generation:
        generation = Generation(config, writer)
        Trace._commit_(writer, Entity.TRACE, trace_id, "add-generation", {
            **generation.data(),
            "id": generation.id,
        })
        return generation

    def retrieval(self, config: RetrievalConfig):
        retrieval = Retrieval(config, self.writer)
        self._commit("add-retrieval", {
            "id": config.id,
            **retrieval.data(),
        })
        return retrieval

    @staticmethod
    def retrieval_(writer: LogWriter, trace_id: str, config: RetrievalConfig):
        retrieval = Retrieval(config, writer)
        Trace._commit_(writer, Entity.TRACE, trace_id, "add-retrieval", {
            "id": config.id,
            **retrieval.data(),
        })
        return retrieval

    def span(self, config: 'SpanConfig') -> 'Span':
        from .span import Span, SpanConfig
        span = Span(config, self.writer)
        self._commit("add-span", {
            "id": config.id,
            **span.data(),
        })
        return span

    @staticmethod
    def span_(writer: LogWriter, trace_id: str, config: 'SpanConfig') -> 'Span':
        from .span import Span, SpanConfig
        span = Span(config, writer)
        Trace._commit_(writer, Entity.TRACE, trace_id, "add-span", {
            "id": config.id,
            **span.data(),
        })
        return span

    def feedback(self, feedback: Feedback):
        self._commit("add-feedback", feedback.__dict__)

    @staticmethod
    def feedback_(writer: LogWriter, trace_id: str, feedback: Feedback):
        Trace._commit_(writer, Entity.TRACE, trace_id,
                       "add-feedback", feedback.__dict__)

    @staticmethod
    def add_tag_(writer: LogWriter, id: str, key: str, value: str):
        EventEmittingBaseContainer._add_tag_(
            writer, Entity.TRACE, id, key, value)

    @staticmethod
    def end_(writer: LogWriter, trace_id: str, data: Optional[Dict[str, str]] = None):
        if data is None:
            data = {}
        return EventEmittingBaseContainer._end_(writer, Entity.TRACE, trace_id, {
            "endTimestamp": datetime.now(),
            **data,
        })

    @staticmethod
    def event_(writer: LogWriter, trace_id: str, id: str, event: str, tags: Optional[Dict[str, str]] = None):
        return EventEmittingBaseContainer._event_(writer, Entity.TRACE, trace_id, id, event, tags)



import logging
import os
import tempfile
import threading
import time
import uuid
from queue import Queue
from typing import Union

from ..apis.maxim import MaximAPI


class LogWriterConfig:
    def __init__(self, base_url, api_key, repository_id, auto_flush=False, flush_interval: Union[int, None] = 10, is_debug=False):
        self.base_url = base_url
        self.api_key = api_key
        self.repository_id = repository_id
        self.auto_flush = auto_flush
        self.flush_interval = flush_interval
        self.is_debug = is_debug


class LogWriter:
    def __init__(self, config: LogWriterConfig):
        self.id = str(uuid.uuid4())
        self.config = config
        self.queue = Queue()
        self.mutex = threading.Lock()
        self.is_debug = config.is_debug
        self.logs_dir = os.path.join(
            tempfile.gettempdir(), f"maxim-sdk/{self.id}/maxim-logs")
        self.__flush_thread = None
        os.makedirs(self.logs_dir, exist_ok=True)
        if self.config.auto_flush:
            if self.config.flush_interval:
                self.__flush_thread = threading.Timer(
                    int(self.config.flush_interval), self.flush)
                self.__flush_thread.start()
            else:
                raise ValueError(
                    "flush_interval is set to None.flush_interval has to be a number")

    def write_to_file(self, logs):
        filename = f"logs-{time.strftime('%Y-%m-%dT%H:%M:%SZ')}.log"
        filepath = os.path.join(self.logs_dir, filename)
        if self.is_debug:
            print(f"Writing logs to file: {filename}")
        with open(filepath, 'w') as file:
            for log in logs:
                file.write(log.serialize() + "\n")
        return filepath

    def flush_log_files(self):
        if os.path.exists(self.logs_dir) == False:
            return
        files = os.listdir(self.logs_dir)
        for file in files:
            with open(os.path.join(self.logs_dir, file), 'r') as f:
                logs = f.read()
            try:
                MaximAPI.pushLogs(
                    self.config.base_url, self.config.api_key, self.config.repository_id, logs)
                os.remove(os.path.join(self.logs_dir, file))
            except Exception as e:
                if self.is_debug:
                    raise Exception(e)

    def flush_logs(self, logs):
        try:
            # Pushing old logs first
            self.flush_log_files()
            # Pushing new logs
            logs_to_push = "\n".join([log.serialize() for log in logs])
            MaximAPI.pushLogs(self.config.base_url, self.config.api_key,
                              self.config.repository_id, logs_to_push)
        except Exception as e:
            self.write_to_file(logs)

    def commit(self, log):
        self.queue.put(log)

    def flush(self):
        with self.mutex:
            items = []
            while not self.queue.empty():
                items.append(self.queue.get())
            if len(items) == 0:
                if self.is_debug:
                    logging.debug("[MaximSDK] No logs to flush")
                return
            if self.is_debug:
                logging.debug("[MaximSDK] Flushing logs to server")
        self.flush_logs(items)
        if self.is_debug:
            print("[MaximSDK] Flush complete")

    def cleanup(self):
        self.flush()
        if self.config.flush_interval and self.__flush_thread:
            self.__flush_thread.cancel()
            self.__flush_thread.join()            
        
